@@ -1241,21 +1241,11 @@ search_func(GtkTreeModel *model, gint column, const gchar *key, GtkTreeIter *ite
 static GtkWidget *
 create_pounces_list(PouncesManager *dialog)
 {
-	GtkWidget *sw;
 	GtkWidget *treeview;
 	GtkTreeSelection *sel;
 	GtkTreeViewColumn *column;
 	GtkCellRenderer *renderer;
 
-	/* Create the scrolled window */
-	sw = gtk_scrolled_window_new(0, 0);
-	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(sw),
-					GTK_POLICY_AUTOMATIC,
-					GTK_POLICY_AUTOMATIC);
-	gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(sw),
-						GTK_SHADOW_IN);
-	gtk_widget_show(sw);
-
 	/* Create the list model */
 	dialog->model = gtk_list_store_new(POUNCES_MANAGER_NUM_COLUMNS,
 									   G_TYPE_POINTER,
@@ -1270,6 +1260,10 @@ create_pounces_list(PouncesManager *dialog)
 #ifdef USE_HILDON
 	if (g_object_class_find_property(g_type_class_peek(GTK_TYPE_TREE_VIEW), "allow-checkbox-mode"))
 		g_object_set (G_OBJECT(treeview), "allow-checkbox-mode", FALSE, NULL);
+#if HILDON_CHECK_VERSION(2,1,0)
+	if (g_object_class_find_property(g_type_class_peek(GTK_TYPE_TREE_VIEW), "hildon-ui-mode"))
+		g_object_set(G_OBJECT(treeview), "hildon-ui-mode", HILDON_UI_MODE_EDIT, NULL);
+#endif /* HILDON > 2.1.0 */
 #endif /* USE_HILDON */
 	g_object_unref(G_OBJECT(dialog->model));
 	dialog->treeview = treeview;
@@ -1283,7 +1277,6 @@ create_pounces_list(PouncesManager *dialog)
 	/* Handle double-clicking */
 	g_signal_connect(G_OBJECT(treeview), "button_press_event",
 					 G_CALLBACK(pounce_double_click_cb), dialog);
-	gtk_container_add(GTK_CONTAINER(sw), treeview);
 	gtk_widget_show(treeview);
 
 	/* Pouncee Column */
@@ -1343,7 +1336,7 @@ create_pounces_list(PouncesManager *dialog)
 	/* Populate list */
 	populate_pounces_list(dialog);
 
-	return sw;
+	return pidgin_make_scrollable(treeview, GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC, GTK_SHADOW_IN, -1, -1);
 }
 
 void
